/*
 * $Id: icmpscan.c,v 1.1.1.1 2006/03/21 23:52:11 dave Exp $
 */
#include "icmpscan.h"
#include "options.h"
#include "targets.h"
#include "packetfu.h"

static	sigjmp_buf	jmpbuf;
static	int		canjump;

/*
 * Signal handler for packet read timeout
 */
void sig_alarm (int sig) {
	if (canjump == 1)
		siglongjmp(jmpbuf, 1);
}

void process_target (struct in_addr *target, int type) {
	struct itimerval	tout, timeroff;
	struct pcap_pkthdr	hdr;
	const  u_char		*rbuf = NULL;
	STATE			state;
	char			*iface;

	logmsg(llINSANE, "Processing %s - %s\n",
		inet_ntoa(*target), pkt_desc(type));

	state.found = 0;
	state.attempt = 0;
	state.target = *target;
	state.type = type;
	state._ip_id = rand() % 0xffff;
	state._icmp_id = rand() % 0xffff;
	state._icmp_seq = rand() % 0xffff;

	memset(&timeroff, 0, sizeof(timeroff));

	if (!(iface = options.interface))
		iface = target_iface(*target);
	if (pcap == NULL || strcmp(pcap_iface, iface) != 0) {
		free(pcap_iface);
		pcap_iface = strdup(iface);
		pcap_init(pcap_iface);
	}

	while (state.attempt++ < options.retries && state.found == 0) {

		memset(&tout, 0, sizeof(tout));
		tout.it_value.tv_sec = (options.timeout / 1000);
		tout.it_value.tv_usec = (options.timeout % 1000) * 1000;

		if (signal(SIGALRM, sig_alarm) == SIG_ERR)
			err(EX_OSERR, "Unable to install signal handler");

		if (sigsetjmp(jmpbuf, 1)) {
			canjump = 0;
			continue;
		}
		canjump = 1;

		send_packet(&state);

		while (1) {
			if (tout.it_value.tv_sec == 0 && tout.it_value.tv_usec == 0) {
				logmsg(llINSANE, "Manually sending SIGALRM\n");
				if (kill(getpid(), SIGALRM) != 0)
					errx(EX_OSERR, "Unable to send SIGALRM");
			}

			logmsg(llINSANE, "Enabling timer (%d.%d)\n",
			    tout.it_value.tv_sec, tout.it_value.tv_usec);
			if (setitimer(ITIMER_REAL, &tout, NULL) != 0)
				err(EX_OSERR, "Unable to set timer!");

			while ((rbuf = pcap_next(pcap, &hdr)) == NULL);

			logmsg(llINSANE, "Received a packet (%d bytes)\n",
				hdr.caplen - datalink_len);

			if (setitimer(ITIMER_REAL, &timeroff, &tout) != 0)
				err(EX_OSERR, "Unable to pause timer");
			logmsg(llINSANE, "Timer paused (%d.%d)\n",
			    tout.it_value.tv_sec, tout.it_value.tv_usec);

			if (process_packet(&state, rbuf, hdr.caplen)) {
				state.found = 1;
				break;
			}
		}
		canjump = 0;
	}

	if (!state.found && options.loglevel >= llVERBOSE)
		printf("%s: -- No response to %s request --\n",
		    inet_ntoa(state.target), pkt_desc(state.type));
}

int
main (int argc, char *argv[]) {
	int		i;
	TARGET_LIST	targets;
	struct in_addr	target;

	/*
	 * Get some entropy
	 */
	srand((unsigned)time(NULL) + getpid());

	process_cmdline(argc, argv);
	network_startup();

	for (i = 0; i < spec_count; i++) {
		if (!init_target_list(&targets, spec_list[i]))
			errx(EX_USAGE, "Bad target specification: %s",
			    target_list_err());

		while (next_target(&targets, &target)) {
			if (options.packets & pktECHO)
				process_target(&target, pktECHO);
			if (options.packets & pktTIMESTAMP)
				process_target(&target, pktTIMESTAMP);
			if (options.packets & pktNETMASK)
				process_target(&target, pktNETMASK);
			if (options.packets & pktINFO)
				process_target(&target, pktINFO);
			if (options.packets & pktROUTER)
				process_target(&target, pktROUTER);
		}
	}

	network_shutdown();

	return(EX_OK);
}
