/*
 * $Id: options.c,v 1.1.1.1 2006/03/21 23:52:11 dave Exp $
 */
#include "options.h"

char	**spec_list = NULL;
int	spec_count = 0;


OPTIONS	options = {
	NULL,			/* Interface */
	0,			/* Promisc */

	{0},			/* Source IP address */
	pktNONE,		/* Packets */
	DEFAULT_TIMEOUT,	/* Timeout */
	DEFAULT_RETRIES,	/* Retries */

	/* Output Options */
	llNORMAL,		/* Log level */
	0,			/* Numeric output */

	/* Internal */
	0,			/* Specified */
};

char	short_options[] =
	"i:cA:"
	"t:r:f:"
	"EPTSNMAIR"
	"hvBn";

struct	option	long_options[] = {
	/* Interface Options */
	{"interface",	1, 0, 'i'},
	{"promisc",	0, 0, 'c'},
	{"address",	1, 0, 'A'},	/* FIXME! */

	{"timeout",	1, 0, 't'},
	{"retries",	1, 0, 'r'},
	{"file",	1, 0, 'f'},

	{"echo",	0, 0, 'E'},
	{"ping"	,	0, 0, 'P'},
	{"timestamp",	0, 0, 'T'},
	{"netmask",	0, 0, 'N'},
	{"info"	,	0, 0, 'I'},
	{"router",	0, 0, 'R'},

	{"help",	0, 0, 'h'},
	{"verbose",	0, 0, 'v'},
	{"debug",	0, 0, 'B'},
	{"numeric",	0, 0, 'n'},

	/* List Terminator */
	{NULL,		0, 0, 0}
};

void usage (char *name) {
	fprintf(stderr, "%s - version %s\n",
	    PACKAGE_NAME, PACKAGE_VERSION);
	fprintf(stderr, "Usage: %s [ options ] target [...]\n", name);
	fprintf(stderr, "  -i <interface>  Specify interface.\n");
	fprintf(stderr, "  -c              Enable promiscuous mode.\n");
	fprintf(stderr, "  -A <address>    Specify source address of generated packets.\n");
	fprintf(stderr, "  -t <timeout>    Specify timeout for probe response.\n");
	fprintf(stderr, "  -r <retries>    Retries per probe.\n");
	fprintf(stderr, "  -f <filename>   Read targets from the specified file.\n");
	fprintf(stderr, "  -E, -P          ICMP Echo Probe\n");
	fprintf(stderr, "  -T, -S          Timestamp\n");
	fprintf(stderr, "  -N, -M          Netmask\n");
	fprintf(stderr, "  -I              Info\n");
	fprintf(stderr, "  -R              Router solicitation\n");
	fprintf(stderr, "  -h              Display usage information\n");
	fprintf(stderr, "  -v              Increase verbosity\n");
	fprintf(stderr, "  -B              Enable debugging output.\n");
	fprintf(stderr, "  -n              Numeric output (do not resolve hostnames)\n");
}

char *
pkt_desc (int type) {
	switch(type) {
	case pktNONE:
		return("None");
	case pktECHO:
		return("Echo");
	case pktTIMESTAMP:
		return("Timestamp");
	case pktNETMASK:
		return("Netmask");
	case pktINFO:
		return("Info");
	case pktROUTER:
		return("Router Solicitation");
	default:
		errx(EX_SOFTWARE, "Unkown packet type in pkt_desc()");
	}
}

int
add_target(char *spec, char **error) {
	TARGET_LIST     target_list;

	if (!init_target_list(&target_list, spec)) {
		*error = target_list_err();
		return(0);
	}

	if ((spec_list = realloc(spec_list, sizeof(char *) * spec_count+1))
	    == NULL)
		err(EX_OSERR, "remalloc() failed");

	if ((spec_list[spec_count] = malloc(strlen(spec)+1)) == NULL)
		err(EX_OSERR, "malloc() failed");
	strcpy(spec_list[spec_count], spec);
	spec_count++;
	*error = NULL;
	return(1);
}

void
process_file(char *filename) {
	FILE	    *infile;
	char	    buffer[BUFFER_MAX_LEN];
	char	    *cp, *tok_store, *error;
	int	     line = 0;

	if (strcasecmp(filename, "-") == 0) {
		if ((infile = fdopen(0, "r")) == NULL)
			err(EX_IOERR, "fdopen() failed");
	} else {
		if ((infile = fopen(filename, "r")) == NULL)
			err(EX_IOERR, "fopen(%s) failed", filename);
	}

	while (fgets(buffer, BUFFER_MAX_LEN-1, infile)) {
		line++;
		cp = strtok_r(buffer, " \t\r\n", &tok_store);
		while (cp != NULL && cp[0] != '#') {
			if (!add_target(cp, &error))
				errx(EX_DATAERR, "Invalid target specification on line %d: %s",
				    line, error);
			cp = strtok_r(NULL, " \t\r\n", &tok_store);
		}
	}

	if (fclose(infile) != 0)
		warn("fclose(%s) failed", filename);
}

void
process_cmdline(int argc, char *argv[]) {
	int		i;
	char		*error;

	if (argc == 1) {
		usage(argv[0]);
		exit(EX_USAGE);
	}

	while ((i = getopt_long(argc, argv, short_options, long_options,
	    NULL)) != -1) {
		switch(i) {
		/* Interface Options */
		case 'i':
			if (options.interface != NULL)
				free(options.interface);
			options.interface = strdup(optarg);
			logmsg(llDEBUG, "Interface = %s\n", options.interface);
			break;
		case 'c':
			options.promisc = 1;
			logmsg(llDEBUG, "Promiscuous mode enabled\n");
			break;
			break;
		case 'A':
			if (!inet_aton(optarg, &options.source_ip))
				errx(EX_USAGE, "Invalid source address");
			options.specified |= SPEC_SRC_ADDR;
			logmsg(llDEBUG, "Source IP address = %s\n",
			    inet_ntoa(options.source_ip));
			break;

		/* */
		case 't':
			if ((options.timeout = atoi(optarg)) == 0)
				errx(EX_USAGE, "Invalid timeout value");
			logmsg(llDEBUG, "Timeout = %d\n", options.timeout);
			break;
		case 'r':
			if ((options.retries = atoi(optarg)) == 0)
				errx(EX_USAGE, "Invalid retries value");
			logmsg(llDEBUG, "Retries = %d\n", options.retries);
			break;
		case 'f':
			process_file(optarg);
			break;

		/* Packet Options */
		case 'E':
		case 'P':
			options.packets |= pktECHO;
			logmsg(llDEBUG, "Enabled Echo requests\n");
			break;
		case 'T':
		case 'S':
			options.packets |= pktTIMESTAMP;
			logmsg(llDEBUG, "Enabled Timestamp requests\n");
			break;
		case 'N':
		case 'M':
			options.packets |= pktNETMASK;
			logmsg(llDEBUG, "Enabled Netmask requests\n");
			break;
		case 'I':
			options.packets |= pktINFO;
			logmsg(llDEBUG, "Enabled Info requests\n");
			break;
		case 'R':
			options.packets |= pktROUTER;
			logmsg(llDEBUG, "Enabled Router Solicitation requests\n");
			break;

		/* Output Options */
		case 'h':
			usage(argv[0]);
			exit(EX_OK);
			/* NOTREACHED */
		case 'v':
			if (++options.loglevel > llINSANE) {
				options.loglevel = llINSANE;
				logmsg(llDEBUG, "Set log level to Insane "
				    "(Too much of a good thing)\n");
			} else
				logmsg(llDEBUG, "Set log level to %s\n",
				    loglevel_desc[options.loglevel]);
			break;
		case 'B':
			options.loglevel = llDEBUG;
			logmsg(llDEBUG, "Set log level to %s\n",
			    loglevel_desc[options.loglevel]);
			break;
		case 'n':
			options.numeric = 1;
			logmsg(llDEBUG, "Numeric output enabled\n");
			break;
		default:
			exit(EX_USAGE);
			/* NOTREACHED */
		}
	}

	/*
	 * Sanity Check options
	 */
	if (options.packets == pktNONE) {
		options.packets = pktECHO | pktTIMESTAMP
		    | pktNETMASK | pktINFO | pktROUTER;
		logmsg(llDEBUG, "Enabled ALL ICMP requests\n");
	}

	/*
	 * Skip of processed options
	 */
	argc -= optind;
	argv += optind;

	/*
	 * Pre-check command line targets
	 */
	for (i = 0; i < argc; i++)
		if (!add_target(argv[i], &error))
			errx(EX_USAGE, "Invalid target specification: %s",
			    error);
}
