/*
 * $Id: packetfu.c,v 1.1.1.1 2006/03/21 23:52:11 dave Exp $
 */
#include "packetfu.h"

intf_t		*intf = NULL;
ip_t		*ip = NULL;
pcap_t		*pcap = NULL;
int		datalink_len = 0;
char		*pcap_iface = NULL;

unsigned char	intf_buffer[1024];
struct intf_entry *intf_info = (struct intf_entry *)intf_buffer;

#define RESET_INTF_INFO \
	memset(intf_info, 0, sizeof(*intf_info)); \
	intf_info->intf_len = sizeof(intf_buffer);

char *
target_iface(struct in_addr addr) {
	struct addr daddr;

	daddr.addr_type = ADDR_TYPE_IP;
	daddr.addr_bits = 32;
	daddr.addr_ip = addr.s_addr;

	RESET_INTF_INFO;
	if (intf_get_dst(intf, intf_info, &daddr) < 0)
		err(EX_OSERR, "intf_get_dst(%s/%d) failed",
		    __FILE__, __LINE__);
	return(intf_info->intf_name);
}

void
pcap_init(char *ifname) {
	char errbuf[PCAP_ERRBUF_SIZE];
	int datalink;
	bpf_u_int32 localnet, netmask;
	struct bpf_program filter_code;
	char pcap_filter[1024];

	/* Close the current pcap handle if there is one */
	if (pcap != NULL)
		pcap_close(pcap);

	/* Get IP Address Of Interface */
	RESET_INTF_INFO;
	strncpy(intf_info->intf_name, ifname, sizeof(intf_info->intf_name));
	if (intf_get(intf, intf_info) < 0)
		err(EX_IOERR, "intf_get(%s:%d) failed", __FILE__, __LINE__);
	memcpy(&options.source_ip, &intf_info->intf_addr.addr_ip,
	    sizeof(options.source_ip));

	/* Create new pcap handle */
	if ((pcap = pcap_open_live(ifname, PCAP_SNAPLEN,
	    options.promisc, PCAP_TIMEOUT, errbuf)) == NULL)
		errx(EX_IOERR, "%s", errbuf);

	/* Calculate datalink header length */
	if ((datalink = pcap_datalink(pcap)) < 0)
		errx(EX_IOERR, "pcap_datalink() failed! (%s)", pcap_geterr(pcap));
	switch(datalink) {
	case DLT_RAW:    datalink_len = 0;  break;
	case DLT_NULL:   datalink_len = 4;  break;
	case DLT_EN10MB: datalink_len = 14; break;
	case DLT_SLIP:   datalink_len = 24; break;
	case DLT_PPP:    datalink_len = 24; break;
	default:
		errx(EX_IOERR, "Unknown data link type! (%d)", datalink);
	}

	/* Get netmask (for pca_compile call later on) */
	if (pcap_lookupnet(ifname, &localnet, &netmask, NULL) < 0)
		errx(EX_IOERR, "pcap_lookupnet failed! (%s)", pcap_geterr(pcap));

	/* Compile our pcap filter */
	snprintf(pcap_filter, sizeof(pcap_filter),
	    PCAP_FILTER, inet_ntoa(options.source_ip));
	if (pcap_compile(pcap, &filter_code, pcap_filter, 1, netmask) < 0)
		errx(EX_IOERR, "pcap_compile failed! (%s)", pcap_geterr(pcap));

	/* Install out pcap filter */
	if (pcap_setfilter(pcap, &filter_code) < 0)
		errx(EX_IOERR, "pcap_setfilter failed! (%s)", pcap_geterr(pcap));

	/* Cleanup */
	pcap_freecode(&filter_code);
}

void
network_startup(void) {
	if ((intf = intf_open()) == NULL)
		err(EX_OSERR, "intf_open(%s:%d) failed", __FILE__, __LINE__);
	if ((ip = ip_open()) == NULL)
		err(EX_OSERR, "ip_open(%s:%d) failed", __FILE__, __LINE__);
}

void
send_packet(STATE *state) {
	unsigned char	buffer[IP_LEN_MAX];
	struct ip_hdr	*iph;
	ICMP_PKT	*icmph;
	time_t		now;
	struct tm	*time_struct;
	int		len;

	memset(buffer, 0x00, sizeof(buffer));
	len = IP_HDR_LEN + ICMP_HDR_LEN + 4;
	iph = (struct ip_hdr *)buffer;
	iph->ip_v = 4;
	iph->ip_hl = 5;
	iph->ip_tos = 0;
	iph->ip_len = 0;
	iph->ip_id = htons(state->_ip_id);
	iph->ip_off = 0;
	iph->ip_ttl = 64;
	iph->ip_p = IPPROTO_ICMP;
	iph->ip_sum = 0;
	iph->ip_src = options.source_ip.s_addr;
	iph->ip_dst = state->target.s_addr;

	icmph = (ICMP_PKT *)(buffer+IP_HDR_LEN);
	icmph->id = state->_icmp_id;
	icmph->seq = state->_icmp_seq;
	switch (state->type) {
	case pktECHO:
		icmph->type = ICMP_ECHO;
		break;
	case pktTIMESTAMP:
		icmph->type = ICMP_TSTAMP;
		len += 12;
		now = time(NULL);
		time_struct = gmtime(&now);
		icmph->payload.timestamp.orig =
			htonl((time_struct->tm_hour * 3600000) +
			(time_struct->tm_min * 60000) +
			(time_struct->tm_sec * 1000));
		break;
	case pktNETMASK:
		icmph->type = ICMP_MASK;
		len += 4;
		break;
	case pktINFO:
		icmph->type = ICMP_INFO;
		break;
	case pktROUTER:
		icmph->type = ICMP_RTRSOLICIT;
		icmph->id = 0;
		icmph->seq = 0;
		break;
	default:
		errx(EX_SOFTWARE, "Invalid state type (%d)", state->type);
		/* NOTREACHED */
	}

	/* Calculate Checksums */
	iph->ip_len = htons(len);
	ip_checksum(buffer, len);

	/* Send Packet */
	if (ip_send(ip, buffer, len) != len) {
		if (errno == EHOSTDOWN)
			state->attempt = options.retries+1;
		else
			err(EX_IOERR, "ip_send() failed");
	}
}

int
process_packet(STATE *state, const u_char *packet, int pktlen) {
	struct ip_hdr	*iph;
	ICMP_PKT	*icmph;

	iph = (struct ip_hdr *)(packet + datalink_len);
	if (iph->ip_p != IPPROTO_ICMP)
		return(0);

	icmph = (ICMP_PKT *)(packet + (iph->ip_hl << 2) + datalink_len);
	switch (icmph->type) {
	case ICMP_ECHOREPLY:
		if (state->type != pktECHO)
			return(0);
		if (icmph->id != state->_icmp_id) {
			logmsg(llDEBUG, "Received ICMP Echo Reply "
			    "with incorrect ICMP ID!\n");
			return(0);
		}
		display_response(&state->target, &iph->ip_src, "Echo");
		return(1);

	case ICMP_TSTAMPREPLY:
		if (state->type != pktTIMESTAMP)
			return(0);

		if (icmph->id != state->_icmp_id) {
			logmsg(llDEBUG, "Received ICMP Timestamp Reply "
			    "with incorrect ICMP ID!\n");
			return(0);
		}

		icmph->ts_orig = ntohl(icmph->ts_orig);
		icmph->ts_rec = ntohl(icmph->ts_rec);
		icmph->ts_trans = ntohl(icmph->ts_trans);

		display_response(&state->target, &iph->ip_src,
		    "Timestamp [0x%08x, 0x%08x, 0x%08x]",
		    icmph->ts_orig, icmph->ts_rec, icmph->ts_trans);
		return(1);

	case ICMP_MASKREPLY:
		if (state->type != pktNETMASK)
			return(0);
		if (icmph->id != state->_icmp_id) {
			logmsg(llDEBUG, "Received ICMP Address Mask reply "
			    "with incorrect ICMP ID!\n");
			return(0);
		}

		display_response(&state->target, &iph->ip_src,
		    "Address Mask [%s]",
		    inet_ntoa(*(struct in_addr *)&icmph->netmask_mask));
		return(1);

	case ICMP_INFOREPLY:
		if (state->type != pktINFO)
			return(0);
		if (icmph->id != state->_icmp_id) {
			logmsg(llDEBUG, "Received ICMP Info "
			    "with incorrect ICMP ID!\n");
			return(0);
		}
		display_response(&state->target, &iph->ip_src, "ICMP Info");
		return(1);

	case ICMP_RTRADVERT:
		if (state->type != pktROUTER)
			return(0);
		if (icmph->id != state->_icmp_id) {
			logmsg(llDEBUG, "Received ICMP Router Advertisement "
			    "with incorrect ICMP ID!\n");
			return(0);
		}
		display_response(&state->target, &iph->ip_src,
		    "Router Advertisement");
		return(1);

	case ICMP_REDIRECT:
		/* FIXME: Do somthing intelligent with redirects */
	default:
		logmsg(llDEBUG, "Unknown ICMP packet from %s (Type=%d, Code=%d)\n",
		    inet_ntoa(*(struct in_addr *)&iph->ip_src),
		    icmph->type, icmph->code);
	}


	return(0);
}

void
network_shutdown(void) {
	if (pcap != NULL)
		pcap_close(pcap);
	if (ip != NULL)
		ip_close(ip);
	if (intf != NULL)
		intf_close(intf);
}
